# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from decimal import Decimal

from trytond.modules.company.tests import CompanyTestMixin
from trytond.modules.currency.tests import create_currency
from trytond.pool import Pool
from trytond.tests.test_tryton import ModuleTestCase, with_transaction
from trytond.transaction import Transaction


class CarrierWeightTestCase(CompanyTestMixin, ModuleTestCase):
    'Test CarrierWeight module'
    module = 'carrier_weight'
    extras = [
        'purchase_shipment_cost', 'sale_shipment_cost', 'stock_shipment_cost']

    def create_carrier(self):
        pool = Pool()
        Party = pool.get('party.party')
        Uom = pool.get('product.uom')
        Template = pool.get('product.template')
        Product = pool.get('product.product')
        Carrier = pool.get('carrier')
        WeightPriceList = pool.get('carrier.weight_price_list')

        party, = Party.create([{
                    'name': 'Carrier',
                    }])
        uom, = Uom.search([
                ('name', '=', 'Unit'),
                ])
        template, = Template.create([{
                    'name': 'Carrier',
                    'default_uom': uom.id,
                    'type': 'service',
                    }])
        product, = Product.create([{
                    'template': template.id,
                    }])
        weight_uom, = Uom.search([
                ('name', '=', 'Kilogram'),
                ])
        currency = create_currency('cu1')
        carrier, = Carrier.create([{
                    'party': party.id,
                    'carrier_product': product.id,
                    'carrier_cost_method': 'weight',
                    'weight_uom': weight_uom.id,
                    'weight_currency': currency.id,
                    }])
        for i, weight in enumerate(range(0, 100, 20), 1):
            WeightPriceList.create([{
                        'carrier': carrier.id,
                        'weight': weight,
                        'price': Decimal(i),
                        }])
        return carrier

    @with_transaction()
    def test_compute_weight_price(self):
        'Test compute_weight_price'
        carrier = self.create_carrier()

        for weight, price in [
                (-1, Decimal(0)),
                (0, Decimal(1)),
                (1, Decimal(1)),
                (10, Decimal(1)),
                (20, Decimal(1)),
                (21, Decimal(2)),
                (80, Decimal(4)),
                (81, Decimal(5)),
                (100, Decimal(5)),
                ]:
            with self.subTest(weight=weight):
                self.assertEqual(carrier.compute_weight_price(weight), price)

    @with_transaction()
    def test_get_weight_price(self):
        "Test get_weight_price"
        transaction = Transaction()
        carrier = self.create_carrier()

        for weights, price in [
                ([], Decimal(1)),
                ([0], Decimal(1)),
                ([0, 0], Decimal(2)),
                ([21], Decimal(2)),
                ([10, 21], Decimal(3)),
                ([0, 21], Decimal(3)),
                ]:
            with self.subTest(weights=weights):
                with transaction.set_context(weights=weights):
                    self.assertEqual(
                        carrier.get_sale_price(),
                        (price, carrier.weight_currency.id))
                    self.assertEqual(
                        carrier.get_purchase_price(),
                        (price, carrier.weight_currency.id))


del ModuleTestCase



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.

from trytond.pool import Pool

from . import carrier, sale, stock


def register():
    Pool.register(
        carrier.Carrier,
        carrier.WeightPriceList,
        module='carrier_weight', type_='model')
    Pool.register(
        stock.ShipmentIn,
        module='carrier_weight', type_='model',
        depends=['purchase_shipment_cost'])
    Pool.register(
        stock.ShipmentOut,
        sale.Sale,
        module='carrier_weight', type_='model',
        depends=['sale_shipment_cost'])



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from decimal import Decimal

from trytond.model import ModelSQL, ModelView, fields
from trytond.modules.currency.fields import Monetary
from trytond.pool import PoolMeta
from trytond.pyson import Bool, Eval, Id
from trytond.transaction import Transaction


class Carrier(metaclass=PoolMeta):
    __name__ = 'carrier'
    weight_uom = fields.Many2One(
        'product.uom', "Weight UoM",
        domain=[('category', '=', Id('product', 'uom_cat_weight'))],
        states={
            'invisible': Eval('carrier_cost_method') != 'weight',
            'required': Eval('carrier_cost_method') == 'weight',
            'readonly': Bool(Eval('weight_price_list', [])),
            },
        help="The Unit of Measure of weight criteria for the price list.")
    weight_currency = fields.Many2One('currency.currency', 'Currency',
        states={
            'invisible': Eval('carrier_cost_method') != 'weight',
            'required': Eval('carrier_cost_method') == 'weight',
            'readonly': Bool(Eval('weight_price_list', [])),
            },
        help="The currency of the price.")
    weight_price_list = fields.One2Many('carrier.weight_price_list', 'carrier',
        'Price List',
        states={
            'invisible': Eval('carrier_cost_method') != 'weight',
            'readonly': ~(Eval('weight_uom', 0) & Eval('weight_currency', 0)),
            },
        help="Add price per weight to the carrier service.\n"
        "The first line for which the weight is greater is used.\n"
        "The line with weight of 0 is used as default price.")

    @classmethod
    def __setup__(cls):
        super(Carrier, cls).__setup__()
        selection = ('weight', 'Weight')
        if selection not in cls.carrier_cost_method.selection:
            cls.carrier_cost_method.selection.append(selection)

    def compute_weight_price(self, weight):
        "Compute price based on weight"
        line = None
        for line in reversed(self.weight_price_list):
            if line.weight < weight:
                return line.price
        else:
            if line and not line.weight and not weight:
                return line.price
        return Decimal(0)

    def _get_weight_price(self):
        weights = Transaction().context.get('weights', [])
        if weights:
            weight_price = sum(
                self.compute_weight_price(w) for w in weights)
        else:
            weight_price = self.compute_weight_price(0)
        return weight_price, self.weight_currency.id

    def get_sale_price(self):
        price, currency_id = super(Carrier, self).get_sale_price()
        if self.carrier_cost_method == 'weight':
            price, currency_id = self._get_weight_price()
        return price, currency_id

    def get_purchase_price(self):
        price, currency_id = super(Carrier, self).get_purchase_price()
        if self.carrier_cost_method == 'weight':
            price, currency_id = self._get_weight_price()
        return price, currency_id


class WeightPriceList(ModelSQL, ModelView):
    'Carrier Weight Price List'
    __name__ = 'carrier.weight_price_list'
    carrier = fields.Many2One(
        'carrier', "Carrier", required=True,
        help="The carrier that the price list belongs to.")
    weight = fields.Float(
        "Weight", digits='weight_uom',
        domain=[
            ('weight', '>=', 0),
            ],
        depends={'carrier'},
        help="The lower limit for the price.")
    price = Monetary(
        "Price", currency='currency', digits='currency',
        help="The price of the carrier service.")

    currency = fields.Function(fields.Many2One(
            'currency.currency', "Currency"),
        'on_change_with_currency')
    weight_uom = fields.Function(fields.Many2One(
            'product.uom', "Weight UoM",
            help="The Unit of Measure of the weight."),
        'on_change_with_weight_uom')

    @classmethod
    def __setup__(cls):
        super(WeightPriceList, cls).__setup__()
        cls._order.insert(0, ('weight', 'ASC'))

    @fields.depends('carrier', '_parent_carrier.weight_currency')
    def on_change_with_currency(self, name=None):
        return self.carrier.weight_currency if self.carrier else None

    @fields.depends('carrier', '_parent_carrier.weight_uom')
    def on_change_with_weight_uom(self, name=None):
        return self.carrier.weight_uom if self.carrier else None



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from functools import partial
from itertools import groupby

from trytond.pool import PoolMeta
from trytond.tools import sortable_values

from .common import parcel_weight


class Sale(metaclass=PoolMeta):
    __name__ = 'sale.sale'

    def _group_parcel_key(self, lines, line):
        """
        The key to group lines by parcel
        """
        return ()

    def _parcel_weight(self, parcel):
        if self.carrier:
            return parcel_weight(parcel, self.carrier.weight_uom, 'unit')

    def _get_carrier_context(self, carrier):
        context = super(Sale, self)._get_carrier_context(carrier)

        if not carrier or carrier.carrier_cost_method != 'weight':
            return context
        context = context.copy()
        weights = []
        context['weights'] = weights

        lines = [l for l in self.lines or [] if l.quantity and l.quantity > 0]
        keyfunc = partial(self._group_parcel_key, lines)
        lines = sorted(lines, key=sortable_values(keyfunc))

        for key, parcel in groupby(lines, key=keyfunc):
            weights.append(self._parcel_weight(parcel))
        return context



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from trytond.pool import Pool


def parcel_weight(parcel, carrier_uom, uom_field='unit'):
    pool = Pool()
    Uom = pool.get('product.uom')

    weight = 0
    for line in parcel:
        product = getattr(line, 'product', None)
        quantity = getattr(line, 'quantity', None)
        uom = getattr(line, uom_field, None)

        if not all([product, quantity, uom]):
            continue

        if product.weight is not None:
            internal_quantity = Uom.compute_qty(
                uom, quantity, product.default_uom, round=False)
            weight += Uom.compute_qty(
                product.weight_uom, internal_quantity * product.weight,
                carrier_uom, round=False)
        elif uom.category == carrier_uom.category:
            weight += Uom.compute_qty(uom, quantity, carrier_uom, round=False)
    return weight



# This file is part of Tryton.  The COPYRIGHT file at the top level of
# this repository contains the full copyright notices and license terms.
from functools import partial
from itertools import groupby

from trytond.model import fields
from trytond.pool import PoolMeta
from trytond.tools import sortable_values

from .common import parcel_weight


class ShipmentIn(metaclass=PoolMeta):
    __name__ = 'stock.shipment.in'

    def _group_parcel_key(self, lines, line):
        """
        The key to group lines by parcel
        """
        return ()

    @fields.depends('carrier')
    def _parcel_weight(self, parcel):
        if self.carrier:
            return parcel_weight(parcel, self.carrier.weight_uom)

    @fields.depends('carrier', 'incoming_moves',
        methods=['_group_parcel_key', '_parcel_weight'])
    def _get_carrier_context(self):
        context = super(ShipmentIn, self)._get_carrier_context()
        if not self.carrier:
            return context
        if self.carrier.carrier_cost_method != 'weight':
            return context
        weights = []
        context['weights'] = weights

        lines = self.incoming_moves or []
        keyfunc = partial(self._group_parcel_key, lines)
        lines = sorted(lines, key=sortable_values(keyfunc))

        for key, parcel in groupby(lines, key=keyfunc):
            weights.append(self._parcel_weight(parcel))
        return context


class ShipmentOut(metaclass=PoolMeta):
    __name__ = 'stock.shipment.out'

    def _group_parcel_key(self, lines, line):
        """
        The key to group lines by parcel
        """
        return ()

    @fields.depends('carrier')
    def _parcel_weight(self, parcel):
        if self.carrier:
            return parcel_weight(parcel, self.carrier.weight_uom)

    @fields.depends(
        'state', 'carrier', 'inventory_moves', 'outgoing_moves',
        'warehouse_storage', 'warehouse_output',
        methods=['_group_parcel_key', '_parcel_weight'])
    def _get_carrier_context(self):
        context = super(ShipmentOut, self)._get_carrier_context()
        if not self.carrier:
            return context
        if self.carrier.carrier_cost_method != 'weight':
            return context
        weights = []
        context['weights'] = weights

        if (self.state in {'draft', 'waiting', 'assigned'}
                and self.warehouse_storage != self.warehouse_output):
            lines = self.inventory_moves or []
        else:
            lines = self.outgoing_moves or []
        keyfunc = partial(self._group_parcel_key, lines)
        lines = sorted(lines, key=sortable_values(keyfunc))

        for key, parcel in groupby(lines, key=keyfunc):
            weights.append(self._parcel_weight(parcel))
        return context
